/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.math.BigDecimal;
import java.util.HashSet;
import uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciiGridImporter;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;
/**
 * A factory for constructing Grid2DSquareCellDouble instances.
 */
public class Grid2DSquareCellDoubleFactory 
        extends Grid2DSquareCellAbstractFactory {
    
    /**
     * The Grid2DSquareCellChunkAbstractFactory for creating chunks.
     */
    protected Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory;
    
    /**
     * The noDataValue for creating chunks.
     */
    protected double noDataValue;
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory.
     * Default:
     * noDataValue to Double.NEGATIVE_INFINITY
     */
    public Grid2DSquareCellDoubleFactory() {
        this( Double.NEGATIVE_INFINITY );
    }
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory.
     * @param noDataValue The noDataValue initially set for construction.
     * Defaults:
     * directory to a new one in System.getProperties( "java.io.tmpdir" );
     */
    public Grid2DSquareCellDoubleFactory(
            double noDataValue ) {
        this(   FileCreator.createTempFile(),
                noDataValue );
        //this ( FileCreator.createNewFile(), noDataValue );
    }
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory.
     * @param directory A "workspace directory" for storing temporary files and
     *   swapping Grid2DSquareCellDouble data to.
     * Defaults:
     * chunkNrows to 64
     * chunkNcols to 64
     * Grid2DSquareCellDoubleChunkFactory to grid2DSquareCellDoubleChunkArray
     */
    public Grid2DSquareCellDoubleFactory(
            File directory ) {
        this(   directory,
                Double.NEGATIVE_INFINITY );
    }
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory.
     * @param directory A "workspace directory" for storing temporary files and
     *   swapping Grid2DSquareCellDouble data to.
     * @param noDataValue The noDataValue initially set for construction.
     * Default:
     * chunkNrows to 64;
     * chunkNcols to 64;
     * Grid2DSquareCellDoubleChunkFactory to grid2DSquareCellDoubleChunkArray.
     */
    public Grid2DSquareCellDoubleFactory(
            File directory,
            double noDataValue ) {
        this(   directory,
                64,
                64,
                new Grid2DSquareCellDoubleChunkArrayFactory(),
                noDataValue );
    }
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory
     * @param directory A "workspace directory" for storing temporary files and
     *   caching Grid2DSquareCellDoubleAbstract data.
     * @param chunkNrows The number of rows chunks have by default.
     * @param chunkNcols The number of columns chunks have by default.
     * @param grid2DSquareCellDoubleChunkFactory The Grid2DSquareCellDoubleChunkAbstractFactory
     *   for creating Grid2DSquareCellDoubleChunks
     * Default:
     * noDataValue to Double.NEGATIVE_INFINITY.
     */
    public Grid2DSquareCellDoubleFactory(
            File directory,
            int chunkNrows,
            int chunkNcols,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory ) {
        this(   directory,
                chunkNrows,
                chunkNcols,
                grid2DSquareCellDoubleChunkFactory,
                Double.NEGATIVE_INFINITY );
    }
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory
     * @param directory A "workspace directory" for storing temporary files and
     *   caching Grid2DSquareCellDoubleAbstract data.
     * @param chunkNrows The number of rows chunks have by default.
     * @param chunkNcols The number of columns chunks have by default.
     * @param grid2DSquareCellDoubleChunkFactory The Grid2DSquareCellDoubleChunkAbstractFactory
     *   for creating Grid2DSquareCellDoubleChunks
     * @param noDataValue The noDataValue initially set for construction.
     */
    public Grid2DSquareCellDoubleFactory(
            File directory,
            int chunkNrows,
            int chunkNcols,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            double noDataValue  ) {
        this.directory = directory;
        boolean isGrid2DSquareCellDoubleChunk64CellMapFactory =
                ( grid2DSquareCellDoubleChunkFactory.getClass() == 
                Grid2DSquareCellDoubleChunk64CellMapFactory.class );
        if ( isGrid2DSquareCellDoubleChunk64CellMapFactory ) {
            if ( chunkNrows * chunkNcols > 64 ) {
                chunkNrows = 8;
                chunkNcols = 8;
            }
        }
        this.directory = directory;
        this.chunkNrows = chunkNrows;
        this.chunkNcols = chunkNcols;
        this.dimensions = new BigDecimal[ 5 ];
        this.dimensions[ 0 ] = new BigDecimal( 1L );
        this.dimensions[ 1 ] = new BigDecimal( 0L );
        this.dimensions[ 2 ] = new BigDecimal( 0L );
        this.dimensions[ 3 ] = new BigDecimal( chunkNcols );
        this.dimensions[ 4 ] = new BigDecimal( chunkNrows );
        this.grid2DSquareCellDoubleChunkFactory = grid2DSquareCellDoubleChunkFactory;
        this.grid2DSquareCells = new HashSet();
        this.gridStatistics = new GridStatistics1();
        this.handleOutOfMemoryError = true;
        this.noDataValue = noDataValue;
    }
    
    /**
     * Returns a reference to this.grid2DSquareCellDoubleChunkFactory.
     */
    public Grid2DSquareCellDoubleChunkAbstractFactory getGrid2DSquareCellDoubleChunkFactory() {
        return this.grid2DSquareCellDoubleChunkFactory;
    }
    
    /**
     * Sets this.grid2DSquareCellDoubleChunkFactory to grid2DSquareCellDoubleChunkFactory.
     */
    public void setGrid2DSquareCellDoubleChunkFactory(
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory ) {
        this.grid2DSquareCellDoubleChunkFactory = grid2DSquareCellDoubleChunkFactory;
    }
    
    /**
     * Returns this.noDataValue.
     */
    public double getNoDataValue() {
        return this.noDataValue;
    }
    
    /**
     * Sets this.noDataValue.
     */
    public void setNoDataValue(
            double noDataValue ) {
        this.noDataValue = noDataValue;
    }
    
    
    /////////////////////////
    // Create from scratch //
    /////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellDouble grid with all values as
     * noDataValues.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param nrows The Grid2DSquareCellDouble nrows.
     * @param ncols The Grid2DSquareCellDouble ncols.
     * @param dimensions
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble create(
            File directory,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return create( this.gridStatistics, 
                directory,
                this.grid2DSquareCellDoubleChunkFactory, 
                nrows,
                ncols,
                dimensions, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble grid with all values as
     * noDataValues.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param nrows The Grid2DSquareCellDouble nrows.
     * @param ncols The Grid2DSquareCellDouble ncols.
     * @param dimensions
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble create(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellDouble(
                gridStatistics,
                directory,
                grid2DSquareCellDoubleChunkFactory,
                this.chunkNrows,
                this.chunkNcols,
                nrows,
                ncols,
                dimensions,
                this.noDataValue,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    
    //////////////////////////////////////////////////////
    // Create from an existing Grid2DSquareCellAbstract //
    //////////////////////////////////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellDouble with all values taken from
     * grid2DSquareCell.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which grid
     *   values are used.
     * @param startRowIndex The topmost row index of grid2DSquareCell.
     * @param startColIndex The leftmost column index of grid2DSquareCell.
     * @param endRowIndex The bottom row index of grid2DSquareCell.
     * @param endColIndex The rightmost column index of grid2DSquareCell.
     * @param grid2DSquareCells A HashSet of swappable Objects of 
     *   Grid2DSquareCellAbstract type.
     * Default:
     * Grid2DSquareCellDoubleChunkAbstractFactory to
     *   this.grid2DSquareCellDoubleChunkFactory
     * GridStatisticsAbstract to GridStatistics1
     */
    public Grid2DSquareCellDouble create(
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return create(
                this.gridStatistics,
                directory,
                grid2DSquareCell,
                this.grid2DSquareCellDoubleChunkFactory,
                startRowIndex,
                startColIndex,
                endRowIndex,
                endColIndex,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with all values taken from
     * grid2DSquareCell.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which grid
     *   values are used.
     * @param startRowIndex The topmost row index of grid2DSquareCell.
     * @param startColIndex The leftmost column index of grid2DSquareCell.
     * @param endRowIndex The bottom row index of grid2DSquareCell.
     * @param endColIndex The rightmost column index of grid2DSquareCell.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble create(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellDouble(
                gridStatistics,
                directory,
                grid2DSquareCell,
                grid2DSquareCellDoubleChunkFactory,
                this.chunkNrows,
                this.chunkNcols,
                startRowIndex,
                startColIndex,
                endRowIndex,
                endColIndex,
                this.noDataValue,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    
    ////////////////////////
    // Create from a File //
    ////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param grid2DSquareCells A HashSet of swappable Objects of 
     *   Grid2DSquareCellAbstract type.
     */
    public Grid2DSquareCellDouble create(
            File directory,
            File gridFile,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return create(
                this.gridStatistics,
                directory,
                gridFile,
                this.grid2DSquareCellDoubleChunkFactory,
                startRowIndex,
                startColIndex,
                endRowIndex,
                endColIndex,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble create(
            GridStatisticsAbstract gridStatistics,
            File directory, File gridFile,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellDouble(
                gridStatistics,
                directory,
                gridFile,
                grid2DSquareCellDoubleChunkFactory,
                this.chunkNrows,
                this.chunkNcols,
                startRowIndex,
                startColIndex,
                endRowIndex,
                endColIndex,
                this.noDataValue,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    /////////////////////////
    // Create from a cache //
    /////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * @param directory The directory for swapping to file.
     * @param gridFile A file containing the data to be used in construction.
     * @param ois The ObjectInputStream to construct from.
     * @param grid2DSquareCells A HashSet of swappable Objects of 
     *   Grid2DSquareCellAbstract type.
     */
    public @Override Grid2DSquareCellDouble create(
            File directory,
            File gridFile,
            ObjectInputStream ois,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellDouble(
                directory,
                gridFile,
                ois,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
}
