/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import gnu.trove.TDoubleHashSet;
import gnu.trove.TIntHashSet;
import java.io.Serializable;
import java.math.BigInteger;
import java.math.BigDecimal;
import uk.ac.leeds.ccg.andyt.ext.math.BigDecimal0;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.CellID;
import uk.ac.leeds.ccg.andyt.grids.core.OutOfMemoryErrorHandler;

/**
 * An abstract class to be extended and methods overridden to provide
 * statistics about the data in Grid2DSquareCellAbstract and
 * Grid2DSquareCellChunkAbstract more optimally.
 */
public abstract class GridStatisticsAbstract
        implements Serializable, GridStatisticsInterface {
    
    private static final long serialVersionUID = 1L;
    
    /**
     * A reference to the Grid2DSquareCellAbstract this is for.
     */
    protected Grid2DSquareCellAbstract grid2DSquareCell;
    
    /**
     * For storing the number of cells with non noDataValues.
     */
    protected BigInteger nonNoDataValueCountBigInteger;
    
    /**
     * For storing the sum of all non noDataValues as a BigDecimal.
     */
    protected BigDecimal sumBigDecimal;
    
    /**
     * For storing the minimum of all non noDataValues as a BigDecimal.
     */
    protected BigDecimal minBigDecimal;
    
    /**
     * For storing the number of min values as a BigInteger.
     */
    protected BigInteger minCountBigInteger;
    
    /**
     * For storing the maximum of all non noDataValues as a BigDecimal.
     */
    protected BigDecimal maxBigDecimal;
    
    /**
     * For storing the number of max values as a BigInteger.
     */
    protected BigInteger maxCountBigInteger;
    
    /**
     * For intitialisation
     */
    protected void init() {
        this.nonNoDataValueCountBigInteger = new BigInteger( "0" );
        this.sumBigDecimal = new BigDecimal( 0.0d );
        //this.minBigDecimal = new BigDecimal( Double.POSITIVE_INFINITY );
        //this.minBigDecimal = new BigDecimal(
        //        "99999999999999999999999999999999999999999" );
        BigDecimal bigDecimalDoubleMAX_VALUE =
                new BigDecimal( Double.MAX_VALUE );
        this.minBigDecimal =
                bigDecimalDoubleMAX_VALUE.add( bigDecimalDoubleMAX_VALUE );
        this.minCountBigInteger = new BigInteger( "0" );
        //this.maxBigDecimal = new BigDecimal( Double.NEGATIVE_INFINITY );
        //this.minBigDecimal = new BigDecimal(
        //        "-99999999999999999999999999999999999999999" );
        BigDecimal bigDecimalDoubleMIN_VALUE =
                new BigDecimal( Double.MIN_VALUE );
        this.maxBigDecimal =
                bigDecimalDoubleMIN_VALUE.add( bigDecimalDoubleMIN_VALUE );
        this.maxCountBigInteger = new BigInteger( "0" );
    }
    
    /**
     * For intitialisation
     */
    protected void init(
            Grid2DSquareCellAbstract grid2DSquareCell ) {
        this.grid2DSquareCell = grid2DSquareCell;
        init();
    }
    
    /**
     * Updates fields from gridStatistics except this.grid2DSquareCell
     * @param gridStatistics the gridStatistics instance which fields are used
     *   to update this.
     */
    protected void update(
            GridStatisticsAbstract gridStatistics ) {
        this.nonNoDataValueCountBigInteger =
                gridStatistics.nonNoDataValueCountBigInteger;
        this.sumBigDecimal = gridStatistics.sumBigDecimal;
        this.minBigDecimal = gridStatistics.minBigDecimal;
        this.minCountBigInteger = gridStatistics.minCountBigInteger;
        this.maxBigDecimal = gridStatistics.maxBigDecimal;
        this.maxCountBigInteger = gridStatistics.maxCountBigInteger;
    }
    
    /**
     * Updates fields (statistics) by going through all values in
     * this.grid2DSquareCellAbstract if they might not be up to date.
     * (NB. After calling this it is inexpensive to convert to GridStatistics0.)
     */
    protected abstract void update();
    
    /**
     * Updates fields (statistics) by going through all values in
     * this.grid2DSquareCellAbstract if they might not be up to date.
     * (NB. After calling this it is inexpensive to convert to GridStatistics0.)
     * This is called from extending classes activated via update. It is here
     * to avoid duplication in the code. The parameters are specified in order
     * to distinguish this method from update().
     * @param nrows The number of rows in the grid.
     * @param ncols The number of columns in the grid.
     */
    protected void update(
            long nrows,
            long ncols ) {
        long row;
        long col;
        if ( this.grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
            //if ( this.grid2DSquareCell instanceof Grid2DSquareCellInt ) {
            Grid2DSquareCellInt grid2DSquareCellInt =
                    ( Grid2DSquareCellInt ) grid2DSquareCell;
            BigDecimal cellBigDecimal;
            int cellInt;
            boolean handleOutOfMemoryError = true;
            int noDataValueInt = grid2DSquareCellInt.getNoDataValue(
                    handleOutOfMemoryError );
            BigDecimal oneBigDecimal = BigDecimal.ONE;
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    cellInt = grid2DSquareCellInt.getCell(
                            row,
                            col,
                            handleOutOfMemoryError );
                    if ( cellInt != noDataValueInt ) {
                        cellBigDecimal = new BigDecimal( cellInt );
                        this.nonNoDataValueCountBigInteger =
                                this.nonNoDataValueCountBigInteger.add( BigInteger.ONE );
                        this.sumBigDecimal = this.sumBigDecimal.add( cellBigDecimal );
                        if ( cellBigDecimal.compareTo( this.minBigDecimal ) == -1 ) {
                            this.minCountBigInteger = BigInteger.ONE;
                            this.minBigDecimal = cellBigDecimal;
                        } else {
                            if ( cellBigDecimal.compareTo( this.minBigDecimal ) == 0 ) {
                                this.minCountBigInteger =
                                        this.minCountBigInteger.add( BigInteger.ONE );
                            }
                        }
                        if ( cellBigDecimal.compareTo( this.maxBigDecimal ) == 1 ) {
                            this.maxCountBigInteger = BigInteger.ONE;
                            this.maxBigDecimal = cellBigDecimal;
                        } else {
                            if ( cellBigDecimal.compareTo( this.maxBigDecimal ) == 0 ) {
                                this.maxCountBigInteger =
                                        this.maxCountBigInteger.add( BigInteger.ONE );
                            }
                        }
                    }
                }
            }
        } else {
            //if ( this.grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class ) {
            //if ( this.grid2DSquareCell instanceof Grid2DSquareCellDouble ) {
            Grid2DSquareCellDouble grid2DSquareCellDouble =
                    ( Grid2DSquareCellDouble ) grid2DSquareCell;
            BigDecimal cellBigDecimal;
            double cellDouble;
            boolean handleOutOfMemoryError = true;
            double noDataValueDouble = grid2DSquareCellDouble.noDataValue;
            BigDecimal oneBigDecimal = BigDecimal.ONE;
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    //System.out.println("row " + row + ", col " + col );
                    cellDouble = grid2DSquareCellDouble.getCell(
                            row,
                            col,
                            handleOutOfMemoryError );
                    if ( cellDouble != noDataValueDouble ) {
                        cellBigDecimal = new BigDecimal( cellDouble );
                        this.nonNoDataValueCountBigInteger =
                                this.nonNoDataValueCountBigInteger.add( BigInteger.ONE );
                        this.sumBigDecimal = this.sumBigDecimal.add( cellBigDecimal );
                        if ( cellBigDecimal.compareTo( this.minBigDecimal ) == -1 ) {
                            this.minCountBigInteger = BigInteger.ONE;
                            this.minBigDecimal = cellBigDecimal;
                        } else {
                            if ( cellBigDecimal.compareTo( this.minBigDecimal ) == 0 ) {
                                this.minCountBigInteger =
                                        this.minCountBigInteger.add( BigInteger.ONE );
                            }
                        }
                        if ( cellBigDecimal.compareTo( this.maxBigDecimal ) == 1 ) {
                            this.maxCountBigInteger = BigInteger.ONE;
                            this.maxBigDecimal = cellBigDecimal;
                        } else {
                            if ( cellBigDecimal.compareTo( this.maxBigDecimal ) == 0 ) {
                                this.maxCountBigInteger =
                                        this.maxCountBigInteger.add( BigInteger.ONE );
                            }
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Returns a String describing this instance
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String toString(
            boolean handleOutOfMemoryError ) {
        try {
            return getDescription();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return toString( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the name of the this class
     */
    protected String getName() {
        return this.getClass().getName();
        //return "GridStatistics0";
        //return "GridStatistics1";
    }
    
    /**
     * Returns a string describing this instance
     */
    protected String getDescription() {
        String result = this.getName();
        result +=
                "( nonNoDataValueCountBigInteger( " +
                this.nonNoDataValueCountBigInteger.toString() + " ), " +
                "maxBigDecimal( " + this.maxBigDecimal.toString() + " ), " +
                "minBigDecimal( " + this.minBigDecimal.toString() + " ), " +
                "maxCountBigInteger( " + this.maxCountBigInteger.toString() + " ), " +
                "minCountBigInteger( " + this.minCountBigInteger.toString() + " ), " +
                "sumBigDecimal( " + this.sumBigDecimal.toString() + " ) )";
        return result;
    }
    
    /**
     * For returning the number of cells with noDataValues as a BigInteger.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigInteger getNonNoDataValueCountBigInteger(
            boolean handleOutOfMemoryError ) {
        try {
            return getNonNoDataValueCountBigInteger();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getNonNoDataValueCountBigInteger( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the number of cells with noDataValues as a BigInteger.
     */
    protected abstract BigInteger getNonNoDataValueCountBigInteger();
    
    /**
     * For returning the number of cells with noDataValues as a long.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getNonNoDataValueCountLong(
            boolean handleOutOfMemoryError ) {
        try {
            return getNonNoDataValueCountLong();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getNonNoDataValueCountLong( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the number of cells with noDataValues as a long.
     */
    protected final long getNonNoDataValueCountLong() {
        boolean handleOutOfMemoryError = true;
        return getNonNoDataValueCountBigInteger( handleOutOfMemoryError ).longValue();
        //BigInteger value;
        //if ( value.compareTo( BigInteger.valueOf( Long.MAX_VALUE ) ) == -1 &&
        //     value.compareTo( BigInteger.valueOf( Long.MIN_VALUE ) ) == 1 ) {
        //    return value.longValue();
        //} else {
        //    if ( valuel.compareTo( BigInteger.valueOf( Long.MAX_VALUE ) ) == 1 ) {
        //        return Long.MAX_VALUE;
        //    } else {
        //        return Long.MIN_VALUE;
        //    }
        //}
    }
    
    /**
     * For returning the number of cells with noDataValues as a int.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getNonNoDataValueCountInt(
            boolean handleOutOfMemoryError ) {
        try {
            return getNonNoDataValueCountInt();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getNonNoDataValueCountInt( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the number of cells with noDataValues as a int.
     */
    protected final int getNonNoDataValueCountInt() {
        boolean handleOutOfMemoryError = true;
        return getNonNoDataValueCountBigInteger( handleOutOfMemoryError ).intValue();
        //BigInteger value;
        //if ( value.compareTo( BigInteger.valueOf( Integer.MAX_VALUE ) ) == -1 &&
        //     value.compareTo( BigInteger.valueOf( Integer.MIN_VALUE ) ) == 1 ) {
        //    return value.intValue();
        //} else {
        //    if ( valuel.compareTo( BigInteger.valueOf( Integer.MAX_VALUE ) ) == 1 ) {
        //        return Integer.MAX_VALUE;
        //    } else {
        //        return Integer.MIN_VALUE;
        //    }
        //}
    }
    
    /**
     * For returning the sum of all non noDataValues as a BigDecimal.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getSumBigDecimal(
            boolean handleOutOfMemoryError ) {
        try {
            return getSumBigDecimal();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getSumBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the sum of all non noDataValues as a BigDecimal.
     */
    protected abstract BigDecimal getSumBigDecimal();
        
    /**
     * For returning the sum of all non noDataValues as a BigInteger.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigInteger getSumBigInteger(
            boolean handleOutOfMemoryError ) {
        try {
            return getSumBigInteger();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getSumBigInteger( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the sum of all non noDataValues as a BigInteger.
     */
    protected final BigInteger getSumBigInteger() {
        boolean handleOutOfMemoryError = true;
        return getSumBigDecimal( handleOutOfMemoryError ).toBigIntegerExact();
    }
    
    /**
     * For returning the sum of all non noDataValues as a double.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getSumDouble(
            boolean handleOutOfMemoryError ) {
        try {
            return getSumDouble();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getSumDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the sum of all non noDataValues as a double.
     * If this.sumBigDecimal is to big/small to represent as a double then an
     * Arithmetic Exception is thrown and Double.POSITIVE_INFINITY or
     * Double.NEGATIVE_INFINITY are returned as appropriate.
     */
    protected final double getSumDouble() {
        boolean handleOutOfMemoryError = true;
        double result = getSumBigDecimal( handleOutOfMemoryError ).doubleValue();
        if ( Double.isInfinite( result ) ) {
            throw new ArithmeticException();
        }
        return result;
    }
    
    /**
     * For returning the sum of all non noDataValues as a long.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getSumLong(
            boolean handleOutOfMemoryError ) {
        try {
            return getSumLong();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getSumLong( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the sum of all non noDataValues as a long.
     */
    protected final long getSumLong() {
        boolean handleOutOfMemoryError = true;
        return getSumBigDecimal( handleOutOfMemoryError ).longValueExact();
        //BigDecimal value;
        //if ( value.compareTo( BigDecimal.valueOf( Long.MAX_VALUE ) ) == -1 &&
        //     value.compareTo( BigDecimal.valueOf( Long.MIN_VALUE ) ) == 1 ) {
        //    return value.longValue();
        //} else {
        //    if ( value.compareTo( BigDecimal.valueOf( Long.MAX_VALUE ) ) == 1 ) {
        //        return Long.MAX_VALUE;
        //    } else {
        //        return Long.MIN_VALUE;
        //    }
        //}
    }
    
    /**
     * For returning the sum of all non noDataValues as a int.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getSumInt(
            boolean handleOutOfMemoryError ) {
        try {
            return getSumInt();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getSumInt( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the sum of all non noDataValues as a int.
     */
    protected int getSumInt() {
        update();
        boolean handleOutOfMemoryError = true;
        return getSumBigDecimal( handleOutOfMemoryError ).intValueExact();
        //BigDecimal value;
        //if ( value.compareTo( BigDecimal.valueOf( Integer.MAX_VALUE ) ) == -1 &&
        //     value.compareTo( BigDecimal.valueOf( Integer.MIN_VALUE ) ) == 1 ) {
        //    return value.intValue();
        //} else {
        //    if ( valuel.compareTo( BigDecimal.valueOf( Integer.MAX_VALUE ) ) == 1 ) {
        //        return Integer.MAX_VALUE;
        //    } else {
        //        return Integer.MIN_VALUE;
        //    }
        //}
    }
    
    /**
     * For returning the minimum of all non noDataValues as a BigDecimal.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getMinBigDecimal(
            boolean handleOutOfMemoryError ) {
        try {
            return getMinBigDecimal();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMinBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the minimum of all non noDataValues as a BigInteger.
     */
    protected abstract BigDecimal getMinBigDecimal();
    
    /**
     * For returning the minimum of all non noDataValues as a BigInteger.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigInteger getMinBigInteger(
            boolean handleOutOfMemoryError ) {
        try {
            return getMinBigInteger();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMinBigInteger( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the minimum of all non noDataValues as a BigInteger.
     */
    protected final BigInteger getMinBigInteger() {
        boolean handleOutOfMemoryError = true;
        return getMinBigDecimal( handleOutOfMemoryError ).toBigIntegerExact();
    }
    
    /**
     * For returning the minimum of all non noDataValues as a double.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getMinDouble(
            boolean handleOutOfMemoryError ) {
        try {
            return getMinDouble();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMinDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the minimum of all non noDataValues as a double.
     * If this.minBigDecimal is to big/small to represent as a double then an
     * Arithmetic Exception is thrown and Double.POSITIVE_INFINITY or
     * Double.NEGATIVE_INFINITY are returned as appropriate.
     */
    protected final double getMinDouble() {
        update();
        double result = this.minBigDecimal.doubleValue();
        if ( Double.isInfinite( result ) ) {
            throw new ArithmeticException();
        }
        return result;
    }
    
    /**
     * For returning the minimum of all non noDataValues as a long.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getMinLong(
            boolean handleOutOfMemoryError ) {
        try {
            return getMinLong();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMinLong( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the minimum of all non noDataValues as a long
     */
    protected final long getMinLong() {
        return getMinBigInteger().longValue();
        //BigDecimal value;
        //if ( value.compareTo( BigDecimal.valueOf( Long.MAX_VALUE ) ) == -1 &&
        //     value.compareTo( BigDecimal.valueOf( Long.MIN_VALUE ) ) == 1 ) {
        //    return value.longValue();
        //} else {
        //    if ( value.compareTo( BigDecimal.valueOf( Long.MAX_VALUE ) ) == 1 ) {
        //        return Long.MAX_VALUE;
        //    } else {
        //        return Long.MIN_VALUE;
        //    }
        //}
    }
    
    /**
     * For returning the minimum of all non noDataValues as a int.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getMinInt(
            boolean handleOutOfMemoryError ) {
        try {
            return getMinInt();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMinInt( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the minimum of all non noDataValues as a int.
     */
    protected final int getMinInt() {
        return getMinBigInteger().intValue();
    }
    
    /**
     * For returning the maximum of all non noDataValues as a BigInteger.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigInteger getMaxBigInteger(
            boolean handleOutOfMemoryError ) {
        try {
            return getMaxBigInteger();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMaxBigInteger( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the maximum of all non noDataValues as a BigInteger.
     */
    protected final BigInteger getMaxBigInteger() {
        boolean handleOutOfMemoryError = true;
        return getMaxBigDecimal( handleOutOfMemoryError ).toBigIntegerExact();
    }
    
    /**
     * For returning the maximum of all non noDataValues as a BigDecimal.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getMaxBigDecimal(
            boolean handleOutOfMemoryError ) {
        try {
            return getMaxBigDecimal();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMaxBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the maximum of all non noDataValues as a BigInteger.
     */
    protected abstract BigDecimal getMaxBigDecimal();
    
    /**
     * For returning the maximum of all non noDataValues as a double.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getMaxDouble(
            boolean handleOutOfMemoryError ) {
        try {
            return getMaxDouble();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMaxDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the maximum of all non noDataValues as a double.
     * If this.minBigDecimal is to big/small to represent as a double then an
     * Arithmetic Exception is thrown and Double.POSITIVE_INFINITY or
     * Double.NEGATIVE_INFINITY are returned as appropriate.
     */
    protected final double getMaxDouble() {
        boolean handleOutOfMemoryError = true;
        double result = getMaxBigDecimal( handleOutOfMemoryError ).doubleValue();
        if ( Double.isInfinite( result ) ) {
            throw new ArithmeticException();
        }
        return result;
    }
    
    /**
     * For returning the maximum of all non noDataValues as a long.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getMaxLong(
            boolean handleOutOfMemoryError ) {
        try {
            return getMaxLong();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMaxLong( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the minimum of all non noDataValues as a long.
     */
    public final long getMaxLong() {
        boolean handleOutOfMemoryError = true;
        return getMaxBigDecimal( handleOutOfMemoryError ).longValueExact();
        //BigDecimal value;
        //if ( value.compareTo( BigDecimal.valueOf( Long.MAX_VALUE ) ) == -1 &&
        //     value.compareTo( BigDecimal.valueOf( Long.MIN_VALUE ) ) == 1 ) {
        //    return value.longValue();
        //} else {
        //    if ( value.compareTo( BigDecimal.valueOf( Long.MAX_VALUE ) ) == 1 ) {
        //        return Long.MAX_VALUE;
        //    } else {
        //        return Long.MIN_VALUE;
        //    }
        //}
    }
    
    /**
     * For returning the maximum of all non noDataValues as a int.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getMaxInt(
            boolean handleOutOfMemoryError ) {
        try {
            return getMaxInt();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getMaxInt( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the minimum of all non noDataValues as a int.
     */
    protected final int getMaxInt() {
        boolean handleOutOfMemoryError = true;
        return getMaxBigDecimal( handleOutOfMemoryError ).intValueExact();
    }
    
    /**
     * For returning the mode of all non noDataValues either as a
     * TDoubleHashSet or as a TIntHashSet respectively depending on
     * if ( this.grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) or
     * if ( this.grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class ).
     * TODO:
     * Change for loops so as to look through each chunk in turn.
     * Is it better to use toArray and go through a sorted version?
     */
    protected Object getMode() {
        long nrows = this.grid2DSquareCell.nrows;
        long ncols = this.grid2DSquareCell.ncols;
        long row;
        long col;
        boolean calculated = false;
        if ( this.grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
            Grid2DSquareCellInt grid2DSquareCellInt =
                    ( Grid2DSquareCellInt ) this.grid2DSquareCell;
            TIntHashSet mode = new TIntHashSet();
            boolean handleOutOfMemoryError = true;
            int noDataValue = grid2DSquareCellInt.getNoDataValue(
                    handleOutOfMemoryError );
            Object[] tmode = initMode(
                    grid2DSquareCellInt,
                    nrows,
                    ncols,
                    noDataValue );
            if ( tmode[0] == null ) {
                return mode;
            } else {
                //double value;
                int value;
                long count;
                long modeCount = ( ( Long ) tmode[ 0 ] ).longValue();
                mode.add( ( ( Integer ) tmode[ 1 ] ).intValue() );
                CellID cellID = ( CellID ) tmode[ 2 ];
                // Do remainder of the row
                row = cellID.cellRowIndex;
                for ( col = cellID.cellColIndex + 1; col < ncols; col ++ ) {
                    value = grid2DSquareCellInt.getCell(
                            row,
                            col,
                            handleOutOfMemoryError );
                    if ( value != noDataValue ) {
                        count = count(
                                grid2DSquareCellInt,
                                row,
                                col,
                                nrows,
                                ncols,
                                value );
                        if ( count > modeCount ) {
                            mode.clear();
                            mode.add( value );
                            modeCount = count;
                        } else {
                            if ( count == modeCount ) {
                                mode.add( value );
                            }
                        }
                    }
                }
                // Do remainder of the grid
                for ( row ++ ; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        value = grid2DSquareCellInt.getCell(
                                row,
                                col,
                                handleOutOfMemoryError );
                        if ( value != noDataValue ) {
                            count = count(
                                    grid2DSquareCellInt,
                                    row,
                                    col,
                                    nrows,
                                    ncols,
                                    value );
                            if ( count > modeCount ) {
                                mode.clear();
                                mode.add( value );
                                modeCount = count;
                            } else {
                                if ( count == modeCount ) {
                                    mode.add( value );
                                }
                            }
                        }
                    }
                }
            }
            return mode;
        } else {
            //this.grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class
            Grid2DSquareCellDouble grid2DSquareCellDouble =
                    ( Grid2DSquareCellDouble ) this.grid2DSquareCell;
            TDoubleHashSet mode = new TDoubleHashSet();
            boolean handleOutOfMemoryError = true;
            double noDataValue = grid2DSquareCellDouble.noDataValue;
            Object[] tmode = initMode(
                    grid2DSquareCellDouble,
                    nrows,
                    ncols,
                    noDataValue );
            if ( tmode[0] == null ) {
                return mode;
            } else {
                double value;
                long count;
                long modeCount = ( ( Long ) tmode[ 0 ] ).longValue();
                mode.add( ( ( Double ) tmode[ 1 ] ).doubleValue() );
                CellID cellID = ( CellID ) tmode[ 2 ];
                // Do remainder of the row
                row = cellID.cellRowIndex;
                for ( col = cellID.cellColIndex + 1; col < ncols; col ++ ) {
                    value = grid2DSquareCellDouble.getCell(
                            row,
                            col,
                            handleOutOfMemoryError );
                    if ( value != noDataValue ) {
                        count = count(
                                grid2DSquareCellDouble,
                                row,
                                col,
                                nrows,
                                ncols,
                                value );
                        if ( count > modeCount ) {
                            mode.clear();
                            mode.add( value );
                            modeCount = count;
                        } else {
                            if ( count == modeCount ) {
                                mode.add( value );
                            }
                        }
                    }
                }
                // Do remainder of the grid
                for ( row ++ ; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        value = grid2DSquareCellDouble.getCell(
                                row,
                                col,
                                handleOutOfMemoryError );
                        if ( value != noDataValue ) {
                            count = count(
                                    grid2DSquareCellDouble,
                                    row,
                                    col,
                                    nrows,
                                    ncols,
                                    value );
                            if ( count > modeCount ) {
                                mode.clear();
                                mode.add( value );
                                modeCount = count;
                            } else {
                                if ( count == modeCount ) {
                                    mode.add( value );
                                }
                            }
                        }
                    }
                }
            }
            return mode;
        }
    }
    
    /**
     * For initialising mode calculation in getMode().
     */
    private Object[] initMode(
            Grid2DSquareCellInt grid2DSquareCellInt,
            long nrows,
            long ncols,
            int noDataValue ) {
        Object[] initMode = new Object[3];
        long modeCount;
        long p;
        long q;
        long row;
        long col;
        int value;
        int thisValue;
        boolean handleOutOfMemoryError = true;
        for ( p = 0; p < nrows; p ++ ) {
            for ( q = 0; q < ncols; q ++ ) {
                value = grid2DSquareCellInt.getCell(
                        p,
                        q,
                        handleOutOfMemoryError );
                if ( value != noDataValue ) {
                    modeCount = 0L;
                    for ( row = 0; row < nrows; row ++ ) {
                        for ( col = 0; col < ncols; col ++ ) {
                            thisValue = grid2DSquareCellInt.getCell(
                                    row,
                                    col,
                                    handleOutOfMemoryError );
                            if ( thisValue == value ) {
                                modeCount ++;
                            }
                        }
                    }
                    initMode[0] = new Long( modeCount );
                    initMode[1] = new Integer( value );
                    initMode[2] = grid2DSquareCellInt.getCellID(
                            p,
                            q,
                            handleOutOfMemoryError );
                    //initMode[2] = new CellID( p, q );
                    return initMode;
                }
            }
        }
        return initMode;
    }
    
    /**
     * For initialising mode calculation in getMode().
     */
    private Object[] initMode(
            Grid2DSquareCellDouble grid2DSquareCellDouble,
            long nrows,
            long ncols,
            double noDataValue ) {
        Object[] initMode = new Object[3];
        long modeCount;
        long p;
        long q;
        long row;
        long col;
        double value;
        double thisValue;
        boolean handleOutOfMemoryError = true;
        for ( p = 0; p < nrows; p ++ ) {
            for ( q = 0; q < ncols; q ++ ) {
                value = grid2DSquareCellDouble.getCell(
                        p,
                        q,
                        handleOutOfMemoryError );
                if ( value != noDataValue ) {
                    modeCount = 0L;
                    for ( row = 0; row < nrows; row ++ ) {
                        for ( col = 0; col < ncols; col ++ ) {
                            thisValue = grid2DSquareCellDouble.getCell(
                                    row,
                                    col,
                                    handleOutOfMemoryError );
                            if ( thisValue == value ) {
                                modeCount ++;
                            }
                        }
                    }
                    initMode[0] = new Long( modeCount );
                    initMode[1] = new Double( value );
                    initMode[2] = grid2DSquareCellDouble.getCellID(
                            p,
                            q,
                            handleOutOfMemoryError );
                    //initMode[2] = new CellID( p, q );
                    return initMode;
                }
            }
        }
        return initMode;
    }
    
    /**
     * Counts the remaining number of values in grid2DSquareCellInt equal to
     * value from cell given by row p and column q counting in row major order.
     * @param row The row index of the cell from which counting starts
     * @param col The column index of the cell from which counting starts
     * @param nrows The number of rows in grid2DSquareCellInt.
     * @param ncols The number of columns in grid2DSquareCellInt.
     * @param value The value to be counted.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public long count(
            Grid2DSquareCellInt grid2DSquareCellInt,
            long row,
            long col,
            long nrows,
            long ncols,
            int value,
            boolean handleOutOfMemoryError ) {
        try {
            return count(
                    grid2DSquareCellInt,
                    row,
                    col,
                    nrows,
                    ncols,
                    value );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return count(
                        grid2DSquareCellInt,
                        row,
                        col,
                        nrows,
                        ncols,
                        value,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Counts the remaining number of values in grid2DSquareCellInt equal to
     * value from cell given by row p and column q counting in row major order.
     * @param row The row index of the cell from which counting starts
     * @param col The column index of the cell from which counting starts
     * @param nrows The number of rows in grid2DSquareCellInt.
     * @param ncols The number of columns in grid2DSquareCellInt.
     * @param value The value to be counted.
     */
    protected static long count(
            Grid2DSquareCellInt grid2DSquareCellInt,
            long row,
            long col,
            long nrows,
            long ncols,
            int value ) {
        long count = 1L;
        int thisValue;
        boolean handleOutOfMemoryError = true;
        // Do remainder of the row of grid2DSquareCellInt
        for ( col ++; col < ncols; col ++ ) {
            thisValue = grid2DSquareCellInt.getCell(
                    row,
                    col,
                    handleOutOfMemoryError );
            if ( thisValue == value ) {
                count ++;
            }
        }
        // Do remainder of the grid2DSquareCellInt
        for ( row ++ ; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                thisValue = grid2DSquareCellInt.getCell(
                        row,
                        col,
                        handleOutOfMemoryError );
                if ( thisValue == value ) {
                    count ++;
                }
            }
        }
        return count;
    }
    
    /**
     * Counts the remaining number of values in grid2DSquareCellDouble equal to
     * value from cell given by row p and column q counting in row major order.
     * @param row The row index of the cell from which counting starts
     * @param col The column index of the cell from which counting starts
     * @param nrows The number of rows in grid2DSquareCellInt.
     * @param ncols The number of columns in grid2DSquareCellInt.
     * @param value The value to be counted.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public long count(
            Grid2DSquareCellDouble grid2DSquareCellDouble,
            long row,
            long col,
            long nrows,
            long ncols,
            double value,
            boolean handleOutOfMemoryError ) {
        try {
            return count(
                    grid2DSquareCellDouble,
                    row,
                    col,
                    nrows,
                    ncols,
                    value );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return count(
                        grid2DSquareCellDouble,
                        row,
                        col,
                        nrows,
                        ncols,
                        value,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Counts the remaining number of values in grid2DSquareCellDouble equal to
     * value from cell given by row p and column q counting in row major order.
     * @param row The row index of the cell from which counting starts
     * @param col The column index of the cell from which counting starts
     * @param nrows The number of rows in grid2DSquareCellInt.
     * @param ncols The number of columns in grid2DSquareCellInt.
     * @param value The value to be counted.
     */
    protected static long count(
            Grid2DSquareCellDouble grid2DSquareCellDouble,
            long row,
            long col,
            long nrows,
            long ncols,
            double value ) {
        long count = 1L;
        double thisValue;
        boolean handleOutOfMemoryError = true;
        // Do remainder of the row of grid2DSquareCellDouble
        for ( col ++; col < ncols; col ++ ) {
            thisValue = grid2DSquareCellDouble.getCell(
                    row,
                    col,
                    handleOutOfMemoryError );
            if ( thisValue == value ) {
                count ++;
            }
        }
        // Do remainder of the grid2DSquareCellDouble
        for ( row ++ ; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                thisValue = grid2DSquareCellDouble.getCell(
                        row,
                        col,
                        handleOutOfMemoryError );
                if ( thisValue == value ) {
                    count ++;
                }
            }
        }
        return count;
    }
    
    /**
     * For returning the arithmetic mean of all non noDataValues as a BigDecimal
     * @param numberOfDecimalPlaces The number of places for which the standard
     *   deviation will be correct.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public BigDecimal getArithmeticMeanBigDecimal(
            int numberOfDecimalPlaces,
            boolean handleOutOfMemoryError ) {
        try {
            return getArithmeticMeanBigDecimal( numberOfDecimalPlaces );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getArithmeticMeanBigDecimal(
                        numberOfDecimalPlaces,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the arithmetic mean of all non noDataValues as a BigDecimal
     * @param numberOfDecimalPlaces The number of decimal places to which 
     *   the result is precise.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are 
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected abstract BigDecimal getArithmeticMeanBigDecimal( 
            int numberOfDecimalPlaces );
    
    
    /**
     * Returns the standard deviation of all non noDataValues as a BigDecimal.
     * @param numberOfDecimalPlaces The number of places for which the standard
     *   deviation will be correct.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public BigDecimal getStandardDeviationBigDecimal(
            int numberOfDecimalPlaces,
            boolean handleOutOfMemoryError ) {
        try {
            return getStandardDeviationBigDecimal( numberOfDecimalPlaces );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCell.clearMemoryReserve();
                this.grid2DSquareCell.swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                this.grid2DSquareCell.initMemoryReserve();
                return getStandardDeviationBigDecimal(
                        numberOfDecimalPlaces,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the standard deviation of all non noDataValues as a double.
     * @param numberOfDecimalPlaces The number of places for which the standard
     *   deviation will be correct.
     * TODO:
     * test
     */
    protected BigDecimal getStandardDeviationBigDecimal(
            int numberOfDecimalPlaces ) {
        boolean handleOutOfMemoryError = true;
        BigDecimal stdev = new BigDecimal( "0" );
        // ( 2 * numberOfDecimalPlaces ) is a guess.
        BigDecimal mean = getArithmeticMeanBigDecimal( 
                2 * numberOfDecimalPlaces, 
                handleOutOfMemoryError );
        int chunkNrows = this.grid2DSquareCell.chunkNrows;
        int chunkNcols = this.grid2DSquareCell.chunkNcols;
        int nChunkRows = this.grid2DSquareCell.getNChunkRows(
                handleOutOfMemoryError );
        int nChunkCols = this.grid2DSquareCell.getNChunkCols(
                handleOutOfMemoryError );
        int chunkRowIndex;
        int chunkColIndex;
        int chunkRow;
        int chunkCol;
        boolean isInGrid;
        if ( this.grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
            Grid2DSquareCellInt grid2DSquareCellInt =
                    ( Grid2DSquareCellInt ) this.grid2DSquareCell;
            int value;
            int noDataValue = grid2DSquareCellInt.getNoDataValue(
                    handleOutOfMemoryError );
            BigDecimal nonNoDataValueCountBigDecimal = new BigDecimal( "-1" );
            BigDecimal differenceFromMean;
            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                    Grid2DSquareCellIntChunkAbstract chunk =
                            grid2DSquareCellInt.getGrid2DSquareCellIntChunk(
                            chunkRowIndex,
                            chunkColIndex,
                            handleOutOfMemoryError );
                    for ( chunkRow = 0; chunkRow < chunkNrows; chunkRow ++ ) {
                        for ( chunkCol = 0; chunkCol < chunkNcols; chunkCol ++ ) {
                            isInGrid = grid2DSquareCellInt.isInGrid(
                                    ( ( long ) chunkRowIndex * ( long ) chunkNrows ) +  ( long ) chunkRow,
                                    ( ( long ) chunkColIndex * ( long ) chunkNcols ) +  ( long ) chunkCol,
                                    handleOutOfMemoryError );
                            if ( isInGrid ) {
                                value = chunk.getCell(
                                        chunkRowIndex,
                                        chunkColIndex,
                                        noDataValue,
                                        handleOutOfMemoryError );
                                if ( value != noDataValue ) {
                                    differenceFromMean = new BigDecimal( value ).subtract( mean );
                                    stdev = stdev.add( differenceFromMean.multiply( differenceFromMean ) );
                                    nonNoDataValueCountBigDecimal =
                                            nonNoDataValueCountBigDecimal.add( BigDecimal.ONE );
                                }
                            }
                        }
                    }
                }
            }
            if ( nonNoDataValueCountBigDecimal.compareTo( BigDecimal.ONE ) != 1 ) {
                return stdev;
            }
            BigDecimal bigDecimal0 = stdev.divide(
                    nonNoDataValueCountBigDecimal,
                    numberOfDecimalPlaces,
                    BigDecimal.ROUND_HALF_EVEN );
            return BigDecimal0.sqrt0(
                    bigDecimal0,
                    numberOfDecimalPlaces );
        } else {
            //this.grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class
            Grid2DSquareCellDouble grid2DSquareCellDouble =
                    ( Grid2DSquareCellDouble ) this.grid2DSquareCell;
            double value;
            double noDataValue = grid2DSquareCellDouble.noDataValue;
            BigDecimal nonNoDataValueCountBigDecimal = new BigDecimal( "-1" );
            BigDecimal differenceFromMean;
            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                    Grid2DSquareCellDoubleChunkAbstract chunk =
                            grid2DSquareCellDouble.getGrid2DSquareCellDoubleChunk(
                            chunkRowIndex,
                            chunkColIndex,
                            handleOutOfMemoryError );
                    for ( chunkRow = 0; chunkRow < chunkNrows; chunkRow ++ ) {
                        for ( chunkCol = 0; chunkCol < chunkNcols; chunkCol ++ ) {
                            isInGrid = grid2DSquareCellDouble.isInGrid(
                                    ( ( long ) chunkRowIndex * ( long ) chunkNrows ) +  ( long ) chunkRow,
                                    ( ( long ) chunkColIndex * ( long ) chunkNcols ) +  ( long ) chunkCol,
                                    handleOutOfMemoryError );
                            if ( isInGrid ) {
                                value = chunk.getCell(
                                        chunkRowIndex,
                                        chunkColIndex,
                                        noDataValue,
                                        handleOutOfMemoryError );
                                if ( value != noDataValue ) {
                                    differenceFromMean = new BigDecimal( value ).subtract( mean );
                                    stdev = stdev.add( differenceFromMean.multiply( differenceFromMean ) );
                                    nonNoDataValueCountBigDecimal =
                                            nonNoDataValueCountBigDecimal.add( BigDecimal.ONE );
                                }
                            }
                        }
                    }
                }
            }
            if ( nonNoDataValueCountBigDecimal.compareTo( BigDecimal.ONE ) != 1 ) {
                return stdev;
            }
            BigDecimal bigDecimal0 = stdev.divide(
                    nonNoDataValueCountBigDecimal,
                    numberOfDecimalPlaces,
                    BigDecimal.ROUND_HALF_EVEN );
            return BigDecimal0.sqrt0(
                    bigDecimal0,
                    numberOfDecimalPlaces );
        }
    }
    
}